# SPDX-FileCopyrightText: 2021 Centrum Wiskune en Informatica
#
# SPDX-License-Identifier: MPL-2.0

from html.parser import HTMLParser
from typing import TYPE_CHECKING, Iterable, Optional, Sequence, Tuple

if TYPE_CHECKING:
    from . import Session


class BatchingIterator:
    def __init__(self, url: str, iterable: Iterable, length: int):
        self.url = url
        self.iterable = iterable
        self.length = length

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.url}>"

    def __len__(self):
        return self.length

    def __iter__(self):
        return self.iterable


class Registry:
    """Dict-implementation of @registry endpoint

    Supports:

    Reading registry records:
    >>> session.registry['plone.app.querystring.field.path.title']
    'Location'

    Updating registry records:
    >>> session.registry['plone.app.querystring.field.path.title'] = 'Value'
    >>> session.registry['plone.app.querystring.field.path.title']
    'Value'

    >>> session.registry['plone.app.querystring.field.path.title'] = 1
    Traceback (most recent call last):
    ...
    ValueError: (1, <type 'unicode'>, 'value')


    See https://plonerestapi.readthedocs.io/en/latest/registry.html
    """

    def __init__(self, session: "Session"):
        self._session = session
        self._url = f"@registry"

    def __getitem__(self, key):
        resp = self._session.get(f"{self._url}/{key}")
        if resp.ok:
            return resp.json()
        error = resp.json()
        if (
            resp.status_code == 503
            and error["type"] == "KeyError"
            and key in error["message"]
        ):
            raise KeyError(key)
        else:
            resp.raise_for_status()

    def __setitem__(self, key, value):
        self.update({key: value})

    def update(self, *args, **kwargs):
        "Update registry like `dict.update` would"
        d = {}
        d.update(*args, **kwargs)
        resp = self._session.patch(f"{self._url}", json=d)
        if resp.ok:
            return
        if resp.status_code == 500:
            error = resp.json()
            if error["type"] == "WrongType":
                exc = ValueError(error["message"])
                exc.response = resp
                raise exc
        resp.raise_for_status()

    def __len__(self) -> int:
        resp = self._session.get(f"{self._url}")
        resp.raise_for_status()
        return resp.json()["items_total"]

    def __iter__(self):
        return iter(
            item["name"] for item in self._session.items(f"{self._url}")
        )



# SPDX-FileCopyrightText: 2021 Centrum Wiskune en Informatica
#
# SPDX-License-Identifier: MPL-2.0

from urllib.parse import urlparse

import pytest
from hypothesis import given
from hypothesis.provisional import urls

from affen import Session, api


@given(urls())
def test_root_always_endswith_single_slash(url):
    plone = Session(api_root=url)
    parts_in = urlparse(url)
    parts_out = urlparse(plone.root)
    assert plone.root.endswith("/")
    assert not plone.root.endswith("//")
    assert parts_out[:2] == parts_in[:2]  # scheme and host are untouched
    assert parts_out[3] == parts_in[3].replace("//", "/")  # path is "the same"


def test_changing_root_unauthenticates(plone):
    assert plone.auth
    assert "admin" in repr(plone)
    plone.root = "https://example.com"
    assert not plone.auth
    assert "Authorization" not in plone.headers
    assert api.ANONYMOUS_USER in repr(plone)


@pytest.mark.skip
@pytest.mark.vcr
def test_changing_root_invalidates_token():
    plone = Session()
    plone.login("admin", "admin")
    assert plone.get("@registry").ok
    old_root = plone.root
    old_token = plone.headers["authorization"]
    # change
    plone.root = "https://example.com"
    # change back
    plone.root = old_root
    plone.headers["authorization"] = old_token
    assert not plone.get("@registry").ok


def test_does_not_unauthenticate_when_root_stays_the_same():
    plone = Session(api_root="https://example.com/")
    plone.auth = ("admin", "admin")
    plone.headers["authorization"] = "Bearer deadbeef1234"

    plone.root = "https://example.com"
    assert plone.auth == ("admin", "admin")
    assert plone.headers["authorization"] == "Bearer deadbeef1234"


def test_repr():
    s = repr(Session())
    assert api.ANONYMOUS_USER in s
    assert "affen" in s
    assert "Session" in s


@pytest.mark.vcr
def test_constructor_authenticates_with_basic_authentication(plone_site):
    plone = Session("admin", "admin")
    assert plone.auth == ("admin", "admin")


def test_repr_shows_user(plone):
    assert "admin" in repr(plone)


@pytest.mark.vcr
def test_iterating_over_folder(plone):
    resp = plone.post(
        "", json={"@type": "Folder", "title": "Folder Iteration"}
    )
    assert resp.ok
    folder_url = resp.json()["@id"]
    for i in range(50):
        page = plone.post(
            folder_url, json={"@type": "Document", "title": f"Page {i} test."}
        )
        assert page.ok, page.json()

    page_titles = [p["title"] for p in plone.items(folder_url)]
    assert set(page_titles) == set([f"Page {i} test." for i in range(50)])
    iterator = plone.items(folder_url)
    assert len(iterator) == 50
    assert "folder-iteration" in repr(iterator)


def test_missing_restapi(vcr):
    with vcr.use_cassette("mixtapes/restapi_not_installed.yaml"):
        with pytest.raises(RuntimeError):
            plone = Session()
            plone.login("admin", "admin")


@pytest.mark.vcr
def test_login_sets_token():
    plone = Session()
    plone.login("admin", "admin")
    assert plone.headers["authorization"].startswith("Bearer")


@pytest.mark.vcr
def test_wrong_credentials():
    with pytest.raises(ValueError) as info:
        Session().login("foo", "bar")


@pytest.mark.vcr
def test_does_not_leak_authentication(plone):
    with pytest.raises(ValueError) as info:
        response = plone.get("https://httpbin.org/headers")
    assert "http://127.0.0.1:8080/Plone" in str(info.value)


@pytest.mark.vcr
def test_accepts_absolute_paths_even_if_api_root_is_not_at_host_root(plone):
    assert plone.get("/@search").ok


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["authorization"],
        "record_mode": "once",
    }



#!/bin/env python3

# SPDX-FileCopyrightText: 2021 Centrum Wiskune en Informatica
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path
from subprocess import run
from time import sleep
from typing import Generator

import pytest
from requests import ConnectionError, get

from affen import Session


def docker_run(host, port, site) -> str:
    proc = run(
        [
            "docker",
            "run",
            "-p",
            f"{host}:{port}:8080",
            "-e",
            f"SITE={site}",
            "--detach",
            "--name",
            "plone_restapi_tests",
            "plone:5-alpine",
        ],
        capture_output=True,
        encoding="utf8",
    )
    if proc.returncode:
        raise RuntimeError("Failed to start Plone instance:\n" + proc.stderr)
    container_id = proc.stdout.strip()
    return "plone_restapi_tests"


def docker_stop(container_id: str) -> None:
    run(["docker", "rm", "-f", container_id])


def is_up(url: str) -> bool:
    try:
        get(url)
    except ConnectionError:
        return False
    else:
        return True


def wait_for_url(
    url: str, timeout: float = 60.0, sleep_time: float = 0.5
) -> None:
    while timeout > 0 and not is_up(url):
        timeout -= sleep_time
        sleep(sleep_time)


@pytest.fixture(scope="session")
def plone_site(pytestconfig) -> Generator[str, None, None]:
    host = "127.0.0.1"
    port = 8080
    site = "Plone"
    url = f"http://{host}:{port}/{site}"
    container_id = None

    # if pytestconfig.getoption("--record-mode") != "none" and not is_up(url):
    if pytestconfig.getoption("--vcr-record") != "none" and not is_up(url):
        container_id = docker_run(host, port, site)
        wait_for_url(url)
    yield url
    if container_id:
        docker_stop(container_id)


@pytest.fixture
def plone(plone_site, vcr):
    "An authenticated Session"
    with vcr.use_cassette("auth.yaml"):
        return Session("admin", "admin", plone_site)


if __name__ == "__main__":
    print("Starting" + docker_run("127.0.0.1", 8080, "Plone"))



# SPDX-FileCopyrightText: 2021 Centrum Wiskune en Informatica
#
# SPDX-License-Identifier: MPL-2.0

import pytest

from affen import Session


@pytest.mark.vcr
def test_registry_raises_keyerror(plone):
    with pytest.raises(KeyError) as info:
        plone.registry["plone-restapi-client-testkey"]
    assert "'plone-restapi-client-testkey'" in str(info.value)


@pytest.mark.vcr
def test_registry_raises_valueerror(plone):
    with pytest.raises(ValueError) as info:
        plone.registry["plone.allow_anon_views_about"] = "Should be bool"


@pytest.mark.vcr
def test_registry_raises_notimplementederror(plone):
    "Creating new keys isn't implemented"
    with pytest.raises(NotImplementedError) as info:
        plone.registry["nuskool"] = "The new oldskool"


@pytest.mark.vcr
def test_registry_implements_len(plone):
    assert len(plone.registry) > 0


@pytest.mark.vcr
def test_registry_implements_iteration(plone):
    for key in plone.registry:
        assert key
        break  # datetime values still throw a 500 error


@pytest.mark.vcr
def test_editing_registry(plone):
    plone.registry["plone.allow_anon_views_about"] = True
    assert plone.registry["plone.allow_anon_views_about"] == True
    plone.registry["plone.allow_anon_views_about"] = False
    assert plone.registry["plone.allow_anon_views_about"] == False


@pytest.mark.vcr
def test_registry_update(plone):
    plone.registry.update(
        {
            "plone.allow_anon_views_about": False,
            "plone.always_show_selector": False,
        }
    )
    assert plone.registry["plone.allow_anon_views_about"] == False
    assert plone.registry["plone.always_show_selector"] == False



# SPDX-FileCopyrightText: 2021 Centrum Wiskune en Informatica
#
# SPDX-License-Identifier: MPL-2.0

import re
from typing import Any, Generator, Union
from urllib.parse import urldefrag, urljoin, urlsplit, urlunsplit

import requests

from .utils import BatchingIterator, Registry

DEFAULT_ROOT = "http://localhost:8080/Plone/"
ANONYMOUS_USER = "Anonymous"


class Session(requests.Session):
    """A `requests.Session` equiped as a plone.restapi client."""

    __api_root = ""

    def __init__(
        self,
        user: str = "",
        password: str = "",
        api_root: str = DEFAULT_ROOT,
    ) -> None:
        """`user` and `password` are used to 'Session.login'
        `api_root` is used as the base for relative urls passed to the CRUD methods.


        >>> request = api.Sssion(api_root='https://example.com').get('@actions').request
        >>> request.url
        'https://example.com/@actions'
        >>> request.headers['Accept']
        'application/json'
        """

        super().__init__()
        self.root = api_root
        self.user = ANONYMOUS_USER
        self.headers.update(accept="application/json")
        self.registry = Registry(self)
        self.hooks["response"] = [self._error_handler]
        if user and password:
            self.auth = (user, password)
            self.user = user

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__}: "
            f"{self.user}>"
        )

    def _error_handler(self, resp: requests.Response, *args, **kwargs):
        if resp.ok:
            return
        if resp.status_code == 500:
            error = resp.json()
            if error["type"] == "NotImplementedError":
                raise NotImplementedError(error["message"])

    @property
    def root(self) -> str:
        return self.__api_root

    @root.setter
    def root(self, url: str):
        url = urlsplit(urldefrag(url)[0]).geturl().strip("/") + "/"
        if self.__api_root != url:
            self.auth = None
            self.headers.pop("Authorization", None)
            self.user = ANONYMOUS_USER
            self.__api_root = url

    def request(
        self,
        method: str,
        url: Any,
        *args,
        **kwargs,
    ) -> requests.Response:
        url = urljoin(self.root, url.lstrip("/"))
        if not url.startswith(self.root):
            raise ValueError(
                f"Making requests to other hosts than {self.root} may leak credentials. "
                "Use a different requests.Session for those or change root"
            )
        return super().request(method, url, *args, **kwargs)

    def login(self, user: str, password: str):
        'Authenticate and add the "Bearer" token to the headers'
        resp = self.post(
            f"@login",
            json={"login": user, "password": password},
            headers={"Accept": "application/json"},
        )
        if not resp.ok:
            error = resp.json()["error"]
            if resp.status_code == 501:
                raise RuntimeError(error)
            if 399 < resp.status_code < 500:
                raise ValueError(error)
            resp.raise_for_status()

        self.user = user
        self.__token = resp.json()["token"]
        self.auth = (user, password)
        self.headers.update(authorization=f"Bearer {self.__token}")
        return resp

    def items(self, container: str) -> BatchingIterator:
        """Iterate over all items in `container`

        Implements an iterable over
        https://plonerestapi.readthedocs.io/en/latest/batching.html

        Implements a cheap __len__ too, so for use in progress bars etc.

        >>> len(session.items('/'))

        NB this is not an atomic operation; ie modifing the number of items in
           `container` during iteration may result in items not yielded.
        """
        resp = self.get(container)
        d = resp.json()
        length = d.get("items_total", len(d.get("items", [])))
        return BatchingIterator(container, self._items(resp), length)

    def _items(
        self, container: Union[str, requests.Response]
    ) -> Generator[dict, None, None]:
        if isinstance(container, requests.Response):
            resp = container
        else:
            resp = self.get(container)
        resp.raise_for_status()
        result = resp.json()
        for item in result.get("items", []):
            yield item

        more = result.get("batching", {}).get("next")
        if more:
            # tail recursion is not a thing in Python
            # this might overflow stack for huge results ?!
            yield from self._items(more)
